/***********************************************************************************
                         anselm.ru [2014-04-28] GNU GPL
***********************************************************************************/
#pragma once

#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/containers/map.hpp>
#include <boost/interprocess/containers/deque.hpp>
#include <boost/interprocess/allocators/allocator.hpp>
#include <boost/interprocess/offset_ptr.hpp>
#include <boost/interprocess/containers/string.hpp>
#include <boost/interprocess/sync/named_condition.hpp> //mutex

#define RESERVE          1024 //8192
#define MAXSIZE_SUFFIX   ".maxsize"
#define MUTEX_SUFFIX     ".mutex"

//typedef boost::interprocess::allocator<void, boost::interprocess::managed_shared_memory::segment_manager> void_allocator;
typedef boost::interprocess::allocator<char, boost::interprocess::managed_shared_memory::segment_manager> char_allocator;
typedef boost::interprocess::basic_string<char, std::char_traits<char>, char_allocator> shm_string;
typedef boost::interprocess::offset_ptr<size_t> psize_t;
//typedef boost::interprocess::allocator<shm_string, boost::interprocess::managed_shared_memory::segment_manager> string_allocator;

template <class KeyType, class ValueType>
class ShareMap;

template <class ElemType>
class ShareDeque;


/*************************************************************************  
                            class Share
*************************************************************************/
class Share : public boost::interprocess::managed_shared_memory {
public:
  Share(const char* name, size_t size) // создание новой памяти
    : boost::interprocess::managed_shared_memory(boost::interprocess::open_or_create, name, size) {} //::create_only
  Share(const char* name) // присоединение к существующей памяти
    : boost::interprocess::managed_shared_memory(boost::interprocess::open_only, name) {}
  
  static void
  remove(const char* name) {
    boost::interprocess::shared_memory_object::remove(name);
    boost::interprocess::named_mutex::remove( (std::string(name) + MUTEX_SUFFIX).c_str() );
    //boost::interprocess::named_mutex::remove("MyMutexCond");
    //boost::interprocess::named_condition::remove("FullCondition");
  }
  
  static bool
  exist(const char* name) {
    bool ret;
    try{
      boost::interprocess::managed_shared_memory(boost::interprocess::open_only, name);
      ret = true;
    }
    catch(...) {
      ret = false;
    }
    return ret;
  }
  
  static bool
  good(const char* name) {
    bool ret;
    try{
      boost::interprocess::managed_shared_memory mem = boost::interprocess::managed_shared_memory(boost::interprocess::open_only, name);
      ret = mem.get_free_memory() < mem.get_size(); // нормальное состояние
    }
    catch(...) {
      ret = false;
    }
    return ret;
  }
  
  
/*************************************************************************  
                constructors/functions of shared objects
*************************************************************************/
  template <class T>  
  T create() {
    //const char_allocator alloc_inst(get_segment_manager());
    //return T(alloc_inst);
    return T(get_segment_manager());
  }
  
  template <class KeyType, class ValueType>
  ShareMap<KeyType, ValueType>
  createShareMap(const char* name, size_t maxsize=0) {
    return ShareMap<KeyType, ValueType>(name, this, maxsize);
  }
  
  template <class KeyType, class ValueType>
  ShareMap<KeyType, ValueType>
  openShareMap(const char* name) {
    return ShareMap<KeyType, ValueType>(name, this);
  }
  
  template <class KeyType, class ValueType>
  bool
  existShareMap(const char* name) {
    return ShareMap<KeyType, ValueType>(name, this).exist(name);
  }
  
  template <class ElemType>
  ShareDeque<ElemType>
  createShareDeque(const char* name, int maxsize=0) {
    return ShareDeque<ElemType>(name, this, maxsize);
  }
  
  template <class ElemType>
  ShareDeque<ElemType>
  openShareDeque(const char* name) {
    return ShareDeque<ElemType>(name, this);
  }
  
  template <class ElemType>
  bool
  existShareDeque(const char* name) {
    return ShareDeque<ElemType>(name, this).exist(name);
  }
  
};



/*************************************************************************  
                            class ShareMap
*************************************************************************/
template <class KeyType, class ValueType>
class ShareMap {
friend class Share;
public:
  typedef boost::interprocess::offset_ptr<ValueType> PValueType;
  typedef std::pair<const KeyType, PValueType> PairType;
  typedef boost::interprocess::allocator<PairType, boost::interprocess::managed_shared_memory::segment_manager> map_allocator;
  typedef boost::interprocess::map<KeyType, PValueType, std::less<KeyType>, map_allocator> Map;
  typedef typename Map::const_iterator const_iterator;
  typedef typename Map::iterator iterator;
  const_iterator   begin() const { return d_map->begin(); }
  const_iterator   end()   const { return d_map->end(); }
    
  bool
  put(const KeyType& id, const ValueType& m) {
    
    if(d_mem->get_free_memory() >= d_mem->get_size()) return false; // такое бывает в случае переполнения
    if(d_mem->get_free_memory() <= sizeof(ValueType)+m.data.size()+RESERVE) return false; //RESERVE - надо иметь запас
    if(max_size()>0 && size()>=max_size()) return false;
    
    d_mutex->lock();
    void* temp1 = d_mem->allocate(sizeof(ValueType)); // если мало памяти, то вызывается исключение
    PValueType p = new (temp1) ValueType(m); //b = static_cast<PValueType>(d_mem->d_segment->allocate(sizeof(ValueType)));
    
    iterator I = d_map->find(id);
    if(I == d_map->end()) {
      d_map->insert( PairType(id, p) ); //(*d_map)[id] = p;
    }
    else { // В случае наличия элемента удаляем его из памяти
      d_mem->deallocate( I->second.get() );
      I->second = p;
    }
    d_mutex->unlock();
    
    return true;
  }
  
/*  bool
  put(const KeyType& id, PValueType pm) {
    if(d_mem->get_free_memory() <= sizeof(ValueType)+pm->data.size()) return false;
    
    iterator I = d_map->find(id);
    
    if(I == d_map->end()) {
      d_map->insert( PairType(id, pm) ); //(*d_map)[id] = pm;
    }
    else {
      d_mem->deallocate( I->second.get() );
      I->second = pm; // ссылка на элемент очереди
      // Хм, хорошо ли такое запутывание очереди и карты?
      // С одной стороны - это ускорение, т.к. не надо пересоздавать
      // элемент. Однако, с другой стороны, в случае крушения очереди,
      // рушится также и карта. С третьей стороны, память всё равно 
      // выделяется в одной куче методом d_mem->allocate.
    }
    
    return true;
  }*/
  
  PValueType
  get(const KeyType& id) {
    return (*d_map)[id];//d_map->find(id)
  }

  size_t
  size() const {// размер очереди
    return d_map->size(); 
  }

  bool
  empty() const {
    return d_map->empty(); 
  }

  void
  free() {
    d_mem->destroy_ptr(d_map); //Destroy map. This will free all contains
    d_mem->destroy<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() );
    boost::interprocess::named_mutex::remove( (d_name + MUTEX_SUFFIX).c_str() );
  }
  
  size_t
  max_size() {
    return *d_max_size;
  }

  bool
  exist() const {
    return d_mem->find<Map>(d_name.c_str()).first && d_mem->find<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() ).first;
  }
  
  const char*
  get_name() const {
    return d_name.c_str();
  }
private:
  ShareMap(const char* name, Share* mem, size_t maxsize) // откроем/создадим
  : d_map(NULL)
  , d_max_size(NULL)
  , d_mem(mem)
  , d_name(name) {
    d_mutex = new boost::interprocess::named_mutex(boost::interprocess::open_or_create, (d_name + MUTEX_SUFFIX).c_str());
    d_mutex->unlock(); //кто создаёт объект, тот имеет право (и должен) освободить мютекс
    
    // откроем/создадим карту
    d_map = d_mem->find<Map>(name).first;
    if(!d_map) {
      const map_allocator alloc_inst( d_mem->get_segment_manager() );
      d_map = d_mem->construct<Map>(name)(std::less<KeyType>(), alloc_inst);
    }
    // откроем/создадим переменную "максимальный размер"
    d_max_size = d_mem->find<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() ).first;
    if(d_max_size)
      *d_max_size = maxsize;
    else
      d_max_size = d_mem->construct<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() )(maxsize);
  }
  
  ShareMap(const char* name, Share* mem) // откроем
  : d_map(NULL)
  , d_max_size(NULL)
  , d_mem(mem)
  , d_name(name) {
    d_mutex = new boost::interprocess::named_mutex(boost::interprocess::open_or_create, (d_name + MUTEX_SUFFIX).c_str());
    
    d_map = d_mem->find<Map>(name).first; // откроем карту    
    d_max_size = d_mem->find<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() ).first; // откроем переменную "максимальный размер"
  }
  
  Map *d_map;
  Share *d_mem;
  psize_t d_max_size;
  std::string d_name;
  boost::interprocess::named_mutex *d_mutex;
};


/*************************************************************************  
                            class ShareDeque
*************************************************************************/
template <class ElemType>
class ShareDeque {
friend class Share;
public:
  typedef boost::interprocess::offset_ptr<ElemType> PElemType;
  typedef boost::interprocess::allocator<PElemType, boost::interprocess::managed_shared_memory::segment_manager> deque_allocator;
  typedef boost::interprocess::deque<PElemType, deque_allocator> Deque;
  typedef typename Deque::const_iterator const_iterator;
  const_iterator   begin() const { return d_deque->begin(); }
  const_iterator   end()   const { return d_deque->end(); }
    
  bool
  push(const ElemType& m) {
    //warning("free=%d <= head=%d + body=%d", d_mem->get_free_memory(), sizeof(ElemType), m.data.size());
    if(d_mem->get_free_memory() >= d_mem->get_size()) return false; // такое бывает в случае переполнения
    if(d_mem->get_free_memory() <= sizeof(ElemType)+m.data.size()+RESERVE) return false; //RESERVE - надо иметь запас
    if(max_size()>0 && size()>=max_size()) return false;
    
    d_mutex->lock();
    void* temp1 = d_mem->allocate(sizeof(ElemType)); // если мало памяти, то вызывается исключение
    PElemType p = new (temp1) ElemType(m); //b = static_cast<PValueType>(d_mem->d_segment->allocate(sizeof(ValueType)));
    d_deque->push_back(p);
    d_mutex->unlock();
    return true;
  }

  PElemType
  pop() {   
    if(d_deque->empty()) return NULL;
    //if(d_deque->size()<2) return NULL;// TODO можно использовать вместо блокировки
    //d_mutex->lock();//TODO надо здесь блоктровку?
    PElemType p = d_deque->front();
    d_deque->pop_front();
    //d_mem->deallocate( p.get() ); //!! элемент надо уничтожать вручную после прочтения
    //d_mutex->unlock();//TODO надо здесь блоктровку?
    return p;
  }

  size_t
  size() const {// размер очереди
    return d_deque->size(); 
  }

  bool
  empty() const {
    return d_deque->empty(); 
  }

  void
  free() {    
    d_mem->destroy_ptr(d_deque); //Destroy d_deque. This will free all contains
    d_mem->destroy<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() );
    boost::interprocess::named_mutex::remove( (d_name + MUTEX_SUFFIX).c_str() );
  }
  
  size_t
  max_size() {
    return *d_max_size;
  }

  bool
  exist() const {
    return d_mem->find<Deque>(d_name.c_str()).first && d_mem->find<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() ).first;
  }
  
  const char*
  get_name() const {
    return d_name.c_str();
  }
private:
  ShareDeque(const char* name, Share* mem, size_t maxsize)// открытие/создание
  : d_deque(NULL)
  , d_max_size(NULL)
  , d_mem(mem)
  , d_name(name) {
    d_mutex = new boost::interprocess::named_mutex(boost::interprocess::open_or_create, (d_name + MUTEX_SUFFIX).c_str());
    d_mutex->unlock(); //кто создаёт объект, тот имеет право (и должен) освободить мютекс
    // откроем/создадим очередь
    d_deque = d_mem->find<Deque>(name).first;
    if(!d_deque) {
      const deque_allocator alloc_inst( d_mem->get_segment_manager() );
      d_deque = d_mem->construct<Deque>(name)(alloc_inst);
    }
    // откроем/создадим переменную "максимальный размер"
    d_max_size = d_mem->find<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() ).first;
    if(d_max_size)
      *d_max_size = maxsize;
    else
      d_max_size = d_mem->construct<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() )(maxsize);
  }
  
  ShareDeque(const char* name, Share* mem) // открытие
  : d_deque(NULL)
  , d_max_size(NULL)
  , d_mem(mem)
  , d_name(name) {
    d_mutex = new boost::interprocess::named_mutex(boost::interprocess::open_or_create, (d_name + MUTEX_SUFFIX).c_str());
    
    d_deque = d_mem->find<Deque>(name).first; // откроем очередь
    d_max_size = d_mem->find<size_t>( (d_name + MAXSIZE_SUFFIX).c_str() ).first;// откроем переменную "максимальный размер"    
  }
  
  Deque *d_deque;
  Share *d_mem;
  psize_t d_max_size;
  std::string d_name;
  boost::interprocess::named_mutex *d_mutex;
};

/*void Share::wait()
{
  boost::interprocess::named_mutex mutex(boost::interprocess::open_or_create,"MyMutexCond");
  boost::interprocess::named_condition full(boost::interprocess::open_or_create,"FullCondition");
  boost::interprocess::scoped_lock<boost::interprocess::named_mutex> lock(mutex, boost::interprocess::try_to_lock);
  full.wait(lock);
}

void Share::notify()
{
  boost::interprocess::named_mutex mutex(boost::interprocess::open_or_create,"MyMutexCond");
  boost::interprocess::named_condition full(boost::interprocess::open_or_create,"FullCondition");
  //boost::interprocess::scoped_lock<boost::interprocess::named_mutex> lock(mutex);
  full.notify_one();
}*/
